(function(exports) {
    const OyaVessel = require('./oya-vessel');
    const Actuator = require('./actuator');
    const Sensor = require('./sensor');
    const Light = require('./light');
    const Switch = require('./switch');
    const Fan = require('./fan');

    // OyaMist bioreactor configuration
    class OyaConf {
        constructor(opts = {}) {
            this.update(opts);
        }
        
        static createVesselConfig(index=0, opts={}) {
            var vessel = new OyaVessel(Object.assign({
                name: `vessel${index+1}`,
            }, opts));
            return vessel.toJSON();
        }

        static get MCU_HAT_NONE() { return { text:"(none)", value:"mcu-hat:none"}; }
        static get TEMP_FAHRENHEIT() { return "F"; }
        static get TEMP_CENTIGRADE() { return "C"; }
        static get EVENT_CYCLE_MIST() { return "event:cycle-mist"; }
        static get EVENT_CYCLE_COOL() { return "event:cycle-cool"; }
        static get EVENT_CYCLE_PRIME() { return "event:cycle-prime"; }
        static get CAMERA_NONE() { return "none"; }
        static get CAMERA_ALWAYS_ON() { return "always-on"; }
        static get CAMERA_MANUAL() { return "manual"; }
        static get CAMERA_WHEN_LIT() { return "when-lit"; }

        update(opts) {
            opts = opts || {};
            var i2cRead = opts.i2cRead;
            var i2cWrite = opts.i2cWrite;

            this.name = opts.name || this.name || 'test';

            this.mcuHat = opts.mcuHat || OyaConf.MCU_HAT_NONE.value;

            { // legacy configuration supported multiple vessels
                this.vessel = this.vessel || OyaConf.createVesselConfig(0);
                opts.vessels && OyaVessel.applyDelta(this.vessel, opts.vessels[0]);
                opts.vessel && OyaVessel.applyDelta(this.vessel, opts.vessel);
            }

            if (this.actuators == null && opts.actuators == null) {
                this.actuators = [];
                this.actuators.push(
                    new Actuator({
                        name: `${Actuator.USAGE_MIST}`,
                        usage: Actuator.USAGE_MIST,
                }));
                this.actuators.push(
                    new Actuator({
                        name: `${Actuator.USAGE_COOL}`,
                        usage: Actuator.USAGE_COOL,
                }));
                this.actuators.push(
                    new Actuator({
                        name: `${Actuator.USAGE_PRIME}`,
                        usage: Actuator.USAGE_PRIME,
                }));
            }
            if (opts.actuators) {
                this.actuators = opts.actuators.map((a,i) => {
                    return new Actuator(a);
                });
            }
            if (opts.sensors) {
                var oldSensors = this.sensors || [];
                this.sensors = opts.sensors.map((newSensor,i) => {
                    return Sensor.update(oldSensors[i], newSensor, {
                        i2cRead,
                        i2cWrite,
                    });
                });
            } else if (this.sensors == null) {
                this.sensors = [];
                var sensorOpts = Object.assign(Sensor.TYPE_NONE, {
                    i2cRead,
                    i2cWrite,
                });
                this.sensors.push(new Sensor(sensorOpts));
                this.sensors.push(new Sensor(sensorOpts));
                this.sensors.push(new Sensor(sensorOpts));
            }
            if (opts.lights) {
                this.lights = opts.lights.map((l,i) => {
                    return new Light(l);
                });
            } else if (this.lights == null) {
                this.lights = [];
                this.lights.push(new Light(Light.LIGHT_FULL));
                this.lights.push(new Light(Light.LIGHT_BLUE));
                this.lights.push(new Light(Light.LIGHT_RED));
            }
            if (opts.switches) {
                this.switches = opts.switches.map((s,i) => {
                    return new Switch(s);
                });
            } else if (this.switches == null) {
                this.switches = [];
                this.switches.push(new Switch({
                    name: 'Prime',
                    event: OyaConf.EVENT_CYCLE_PRIME,
                }));
                this.switches.push(new Switch({
                    name: 'Cool',
                    event: OyaConf.EVENT_CYCLE_COOL,
                }));
                this.switches.push(new Switch({
                    name: 'Mist',
                    event: OyaConf.EVENT_CYCLE_MIST,
                }));
            }
            this.fan = this.fan || new Fan();
            this.fan.update(opts.fan);

            this.tempUnit = opts.tempUnit || this.tempUnit || OyaConf.TEMP_FAHRENHEIT;
            this.mcuHat = opts.mcuHat || this.mcuHat || OyaConf.MCU_HAT_NONE.value;
            this.hostTimeout = opts.hostTimeout == null ? 200 : opts.hostTimeout;
            this.healthPoll = opts.healthPoll || 60;
            this.chart = opts.chart || {};
            this.chart.ecStepSize = this.chart.ecStepSize || 10;
            this.chart.tempStepSize = this.chart.tempStepSize || 2;
            this.chart.humidityStepSize = this.chart.humidityStepSize || 5;
            this.chart.showRaw = this.chart.showRaw == null ? false : opts.chart.showRaw;
            this.camera = opts.camera || this.camera || OyaConf.CAMERA_NONE;
            this.camera === true && (this.camera = OyaConf.CAMERA_MANUAL);
            this.heapReboot = opts.heapReboot || 60*1000*1000;

            return this;
        }

        sensorOfField(field) {
            return this.sensors.filter(s => s.isFieldSource(field))[0];
        }

        toJSON() {
            return {
                name: this.name,
                type: "OyaConf",
                tempUnit: this.tempUnit,
                vessel: this.vessel,
                actuators: this.actuators,
                sensors: this.sensors,
                lights: this.lights,
                switches: this.switches,
                mcuHat: this.mcuHat,
                hostTimeout: this.hostTimeout,
                healthPoll: this.healthPoll,
                chart: this.chart,
                camera: this.camera,
                heapReboot: this.heapReboot,
                fan: this.fan,
            };
        }

    } //// class OyaConf

    module.exports = exports.OyaConf = OyaConf;
})(typeof exports === "object" ? exports : (exports = {}));

